# type: ignore
import pytest

from boiler.definitions import ActivityType, ActorType
from boiler.models import Activity, Actor, Box, Detection
from boiler.serialization import kpf


@pytest.mark.skip('Fix immutable models')
def test_serialize_types():
    actor1 = Actor(clip_id=1, actor_type=ActorType.VEHICLE)
    actor2 = Actor(clip_id=2, actor_type=ActorType.PERSON)
    actor3 = Actor(clip_id=3, actor_type=ActorType.PERSON)
    string_gen = kpf.serialize_types([actor1, actor2, actor3])
    produced = [s for s in string_gen]
    expected = [
        '- {types: {cset3: {Vehicle: 1.0}, id1: 1}}\n',
        '- {types: {cset3: {Person: 1.0}, id1: 2}}\n',
        '- {types: {cset3: {Person: 1.0}, id1: 3}}\n',
    ]
    assert produced == expected


@pytest.mark.skip('Fix immutable models')
def test_serialize_geom():
    b1 = Box(left=0, top=1, right=2, bottom=3)
    b2 = Box(left=2, top=4, right=6, bottom=8)
    b3 = Box(left=5, top=4, right=7, bottom=8)
    d1 = Detection(frame=0, box=b1, keyframe=True)
    d2 = Detection(frame=1, box=b2, keyframe=False)
    d3 = Detection(frame=2, box=b3, keyframe=True)
    actor1 = Actor(clip_id=1, actor_type=ActorType.VEHICLE, detections=[d1, d2, d3])
    actor2 = Actor(clip_id=2, actor_type=ActorType.PERSON, detections=[d1, d3])
    string_gen = kpf.serialize_geom([actor1, actor2])
    produced = [s for s in string_gen]
    expected = [
        '- {geom: {g0: 0 1 2 3, id0: 1, id1: 1, keyframe: true, ts0: 0}}\n',
        '- {geom: {g0: 2 4 6 8, id0: 2, id1: 1, ts0: 1}}\n',
        '- {geom: {g0: 5 4 7 8, id0: 3, id1: 1, keyframe: true, ts0: 2}}\n',
        '- {geom: {g0: 0 1 2 3, id0: 4, id1: 2, keyframe: true, ts0: 0}}\n',
        '- {geom: {g0: 5 4 7 8, id0: 5, id1: 2, keyframe: true, ts0: 2}}\n',
    ]
    assert produced == expected


@pytest.mark.skip('Fix immutable models')
def test_serialize_activities_no_id():
    actor1 = Actor(begin=0, end=2)
    actor2 = Actor(begin=1, end=2)
    actor3 = Actor(begin=3, end=4)
    activity1 = Activity(
        activity_type=ActivityType.PEOPLE_TALKING, begin=0, end=2, actors=[actor1, actor2],
    )
    activity2 = Activity(
        activity_type=ActivityType.PURCHASING, begin=3, end=4, status='foo', actors=[actor3],
    )
    # activity_map, _ = assign_entity_ids([activity1, activity2])
    activity_map = {0: activity1, 1: activity2}
    string_gen = kpf.serialize_activities(activity_map.values())
    produced = [s for s in string_gen]
    expected = [
        '- {act: {act2: {People_Talking: 1.0}, actors: [{id1: 1, timespan: [{tsr0: [0, 2]}]}, {id1: 2, timespan: [{tsr0: [1, 2]}]}], id2: 1, timespan: [{tsr0: [0, 2]}]}}\n',  # noqa: E501
        '- {act: {act2: {Purchasing: 1.0}, actors: [{id1: 3, timespan: [{tsr0: [3, 4]}]}], id2: 2, src_status: foo, timespan: [{tsr0: [3, 4]}]}}\n',  # noqa: E501
    ]
    assert produced == expected



"""
python object serialization for boiler models, mostly for debugging and
console logging purposes.  If you find yourself writing tools to parse
these, we should stop and evaluate why.
"""
from typing import Any, Callable, Dict, List

from boiler.models import Activity, Actor, Box, Detection


def deserialize_keyframes(detections: List[Dict[str, Any]]) -> List[Detection]:
    keyframes = [deserialize_detection(d) for d in detections if d['keyframe']]
    return sorted(keyframes, key=lambda d: d.frame)


def deserialize_detection(detection: Dict[str, Any]) -> Detection:
    box = Box(
        left=detection['left'],
        right=detection['right'],
        top=detection['top'],
        bottom=detection['bottom'],
    )
    return Detection(frame=detection['frame'], box=box, keyframe=detection['keyframe'])


def serialize_detection(detection: Detection):
    d = {
        'frame': detection.frame,
        'top': detection.box.top,
        'left': detection.box.left,
        'bottom': detection.box.bottom,
        'right': detection.box.right,
    }
    if detection.keyframe:
        d['keyframe'] = True
    return d


# DANGER: this function cannot be changed without a json-sche
def serialize_actor(actor: Actor, detection_serializer: Callable = serialize_detection):
    return {
        'type': actor.actor_type,
        'range': [actor.begin, actor.end],
        'detections': [detection_serializer(d) for d in actor.detections],
    }


def serialize_activity(
    activity: Activity,
    actor_serializer: Callable = serialize_actor,
    detection_serializer: Callable = serialize_detection,
):
    return {
        'type': activity.activity_type,
        'range': [activity.begin, activity.end],
        'actors': [
            actor_serializer(a, detection_serializer=serialize_detection) for a in activity.actors
        ],
    }


def _serialize_error(err):
    e = err
    ret = []
    while e:
        ret.append(f': {e}')
        e = e.__cause__
    return ret


def serialize_validation_errors(fatals):
    output = []
    for fatal in fatals:
        output.append(_serialize_error(fatal))
    return [' '.join(errs) for errs in output]



import os
import webbrowser

import click
import requests
from xdg import BaseDirectory

from boiler import BOILER_CONFIG_PATH, BOILER_CREDENTIAL_FILE, cli


def _is_valid_stumpf_token(stumpf_url, token):
    # note - don't use the boiler session since that could have an old token,
    # and automatically fails on auth failures.
    return requests.get(stumpf_url + '/api/v1/user/me', headers={'X-Stumpf-Token': token}).ok


@cli.command(name='login')
@click.pass_obj
def login(ctx):
    click.echo('A browser window has been opened, login and copy the token to login.', err=True)
    webbrowser.open(ctx['stumpf_url'] + '/login?next=/token')

    while True:
        stumpf_token = click.prompt('Token', err=True)
        if _is_valid_stumpf_token(ctx['stumpf_url'], stumpf_token):
            with open(
                os.path.join(
                    BaseDirectory.load_first_config(BOILER_CONFIG_PATH), BOILER_CREDENTIAL_FILE
                ),
                'w',
            ) as outfile:
                outfile.write(stumpf_token)
            return click.echo(click.style('You are now logged in.', fg='green'), err=True)
        else:
            click.echo(
                click.style(
                    "Your token doesn't appear to be valid, did you copy/paste correctly?",
                    fg='yellow',
                ),
                err=True,
            )



from pathlib import Path
import re
from typing import Any, Dict, Iterator, List, Set, TextIO, Tuple

import click
from git import InvalidGitRepositoryError, Repo

from boiler import cli
from boiler.commands.utils import activity_types_from_file, summarize_activities
from boiler.serialization.kw18 import KW18FileSet
from boiler.serialization.web import serialize_validation_errors
from boiler.validate import validate_activities

_video_name_regexp = re.compile(
    r'(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})'
    r'\.(?P<start_hour>\d{2})-(?P<start_minute>\d{2})-(?P<start_second>\d{2})'
    r'\.(?P<end_hour>\d{2})-(?P<end_minute>\d{2})-(?P<end_second>\d{2})'
    r'\.(?P<location>[^./]+)'
    r'\.(?P<gtag>G[^./]+)'
)
GENERIC_ERROR = 'Something went wrong communicating with the server...'


def _find_kw18_files(paths: List[str], recursive: bool) -> Iterator[Path]:
    if recursive:
        for p in paths:
            for g in Path(p).glob('**/*.kw18'):
                yield g.parent
    else:
        for p in paths:
            yield Path(p)


def _validate_kw18(path: Path) -> Tuple[List[str], List[str], Dict[str, Any]]:
    kw18_file_set = KW18FileSet.create_from_path(path)
    try:
        activity_list = kw18_file_set.read()
    except Exception as err:
        return [], serialize_validation_errors([err]), {}

    summary = summarize_activities(activity_list)
    warnings, errors = validate_activities(activity_list)
    error = serialize_validation_errors(errors)
    warning = serialize_validation_errors(warnings)
    return warning, error, summary


def get_suffix(count: int, singular: str, plural: str) -> str:
    if count == 1:
        return singular
    return plural


def _get_repo_path(path: Path) -> str:
    """Return the path relative to the git repo root directory."""
    path = path.absolute()
    try:
        repo = Repo(path, search_parent_directories=True)
    except InvalidGitRepositoryError:
        click.echo('Could not find the git repository.  Is this file inside the annotation repo?')
        raise

    return str(path.relative_to(repo.working_tree_dir))


def _serialize_summary(summary: Dict[str, Any], indent: str) -> List[str]:
    default_dict = {'count': 0, 'by_type': []}
    activities = summary.get('activities', default_dict)
    actors = summary.get('actors', default_dict)

    content: List[str] = []

    count = activities['count']
    a = get_suffix(count, 'activity', 'activities')
    content.append(f'{indent}{count} {a}')
    for activity in activities['by_type']:
        name = activity['name']
        count = activity['count']
        content.append(f'{indent}  * {name} ({count})')

    count = actors['count']
    a = get_suffix(count, 'actor', 'actors')
    content.append(f'{indent}{count} {a}')
    for actor in actors['by_type']:
        name = actor['name']
        count = actor['count']
        content.append(f'{indent}  * {name} ({count})')

    return content


def _serialize_output(
    path: Path, error: List[str], warning: List[str], summary: Dict[str, Any],
) -> str:
    prefix = click.style(str(path), bold=True)
    content: List[str] = []
    if error:
        mark = click.style(f'✗  {prefix}', fg='red')
    elif warning:
        mark = click.style(f'⚠  {prefix}', fg='yellow')
    else:
        mark = click.style(f'✔️  {prefix}', fg='green')

    content.extend(
        [
            # str(path),  TODO: maybe add a verbose option for full path
            mark
        ]
    )
    indent = ' ' * 4

    for e in error:
        content.append(click.style(f'{indent}{e}', fg='red'))

    for w in warning:
        content.append(click.style(f'{indent}{w}', fg='yellow'))

    content.extend(_serialize_summary(summary, indent))

    return '\n'.join(content)


@click.group(name='kw18', short_help='Ingest and validate kw18')
@click.pass_obj
def kw18(ctx):
    pass


@kw18.command(name='validate', help='validate one or more kw18 annotations')
@click.argument('path', type=click.Path(exists=True, file_okay=False, resolve_path=True), nargs=-1)
@click.option('--recursive', '-r', is_flag=True)
@click.option('--error-only', is_flag=True, help='Only output information on validation failure')
@click.pass_context
def validate(ctx, path: List[str], recursive: bool, error_only: bool):
    has_invalid = False
    for base_path in _find_kw18_files(path, recursive):
        try:
            warning, error, summary = _validate_kw18(base_path)
        except FileNotFoundError:
            warning = []
            summary = {}
            error = ['Could not determine kw18 file']

        try:
            p = next(base_path.glob('*.kw18'))
        except StopIteration:
            p = base_path.absolute()

        if error_only:
            warning = []

        if error or not error_only:
            click.echo(_serialize_output(path=p, error=error, warning=warning, summary=summary))
            click.echo('')

    ctx.exit(0 if not has_invalid else 1)


@kw18.command(name='dispatch', short_help='dispatch annotation tasks from existing kw18 files')
@click.argument(
    'path',
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    nargs=-1,
    required=True,
)
@click.option('--recursive', '-r', is_flag=True)
@click.option('--vendor', type=click.STRING, required=True, help='the annotation vendor')
@click.option(
    '--activity-type-list',
    type=click.File(mode='r'),
    required=True,
    help='the activity types annotated',
)
@click.option('--set-name', type=click.STRING, required=True)
@click.pass_context
def dispatch(
    ctx, path: List[str], recursive: bool, vendor: str, activity_type_list: TextIO, set_name: str
):
    session = ctx.obj['session']
    has_error = False
    activity_types = activity_types_from_file(activity_type_list)
    for base_path in _find_kw18_files(path, recursive):
        video_name_match = _video_name_regexp.search(str(base_path))
        if video_name_match is None:
            has_error = True
            click.echo(f'Could not determine video name from "{base_path}"')
            continue

        repo_path = _get_repo_path(base_path)
        data = {
            'vendor_name': vendor,
            'activity_types': activity_types,
            'set_name': set_name,
            'annotation_repo_path': repo_path,
            'video_name': video_name_match.group(),
        }
        r = session.post('video-pipeline/annotate', json=data)
        try:
            response_data = r.json()
        except Exception:
            response_data = {}

        if r.status_code == 201:
            click.secho(f'Dispatched {repo_path}', fg='green')
        else:
            message = response_data.get('error', GENERIC_ERROR)
            has_error = True
            click.secho(f'Error dispatching {repo_path}', fg='red')
            click.echo(f'=> {message}')
        click.echo('')

    ctx.exit(0 if not has_error else 1)


@kw18.command(name='ingest', short_help='push files from the annotation repository')
@click.argument(
    'path',
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    nargs=-1,
    required=True,
)
@click.option('--recursive', '-r', is_flag=True)
@click.pass_context
def ingest(ctx, path, recursive):
    """Ingest one or more annotations from the annotation repository.

    This command takes a set of KW18 annotation files from the provided paths
    and sends them to Stumpf.  Stumpf will first detect whether the files have
    changed or not.  If they have not, no further action will be taken.  If
    they have, then Stumpf will:

    \b
    1. Generate a transition to the "annotation" status
    2. Run server side validation.
    3. If validation fails, return failure information.
    4. If validation succeeds, transition to the "audit" state
       and ingest activities from the KW18 files.
    """
    session = ctx.obj['session']
    for base_path in _find_kw18_files(path, recursive):
        repo_path = _get_repo_path(base_path)
        try:
            warning, error, summary = _validate_kw18(base_path)
        except FileNotFoundError:
            warning = []
            summary = {}
            error = ['Could not determine kw18 file']

        click.echo(
            _serialize_output(path=Path(repo_path), error=error, warning=warning, summary=summary)
        )
        click.echo('')

        kw18_file_upload = KW18FileSet.create_from_path(base_path).upload(session)
        repo_path = _get_repo_path(base_path)

        data = {
            'annotation_repo_path': repo_path,
            'kw18_file_upload_id': kw18_file_upload['id'],
        }

        r = session.post('video-pipeline/process-annotation', json=data)
        try:
            response_data = r.json()
        except Exception:
            response_data = {}

        if not r.ok:
            message = response_data.get('error', GENERIC_ERROR)
            click.secho('Error uploading annotations', fg='red')
            click.echo(f'=> {message}')
            click.echo('')
            continue

        status_changes = response_data.get('status_changes')
        if status_changes:
            recorded: Set[str] = set()
            click.echo('Status changes:')
            for sc in status_changes:
                sc_formatted = f'{sc["old_status"]} => {sc["new_status"]}'
                if sc_formatted not in recorded:
                    click.echo(f'    * {sc_formatted}')
                    recorded.add(sc_formatted)
        else:
            click.echo('  No file changes detected')

        click.echo('')


cli.add_command(kw18)



from pathlib import Path
from typing import List

import click

from boiler import cli


@click.group('export', short_help='save current annotation data as kpf files')
def export():
    pass


@export.command(name='video', short_help='export data by video name')
@click.argument('video', nargs=-1)
@click.option(
    '--output', type=click.Path(exists=True, file_okay=False, resolve_path=True), default='.'
)
@click.option('--force/--no-force', help='overwrite existing files', default=False)
@click.pass_context
def video(ctx, output: str, video: List[str], force: bool):
    session = ctx.obj['session']
    success = True
    for video_name in video:
        resp = session.get('video', params={'name': video_name})
        output_path = Path(output) / f'{video_name}.zip'

        if output_path.exists() and not force:
            click.echo(f'A file already exists at {output_path}')
            success = False
            continue

        if resp.status_code != 200 or not resp.json():
            click.echo(f'Could not find {video_name}')
            success = False
            continue

        video_id = resp.json()[0]['id']

        click.echo(f'Downloading {output_path}...')
        resp = session.get(f'video/{video_id}/kpf')
        if resp.status_code != 200:
            click.echo(f'Error downloading {video_name}')
            success = False
            continue

        with output_path.open('wb') as f:
            f.write(resp.content)

    ctx.exit(0 if success else 1)


cli.add_command(export)



from pathlib import Path
import sys

import click
from tabulate import tabulate

from boiler import BoilerSession, cli
from boiler.commands.utils import exit_with, handle_request_error
from boiler.definitions import CameraLocation, ReleaseBatches, Scenarios


def ingest_video(video_path: str, release_batch: str, session: BoilerSession):
    video_name = Path(video_path).stem

    r = session.get('video', json={'name': video_name})
    out = handle_request_error(r)
    if r.ok and len(out['response']):
        click.echo(f'name={video_name} already exists', err=True)
        return out
    elif not r.ok:
        click.echo(f'Could not query server for video={video_name}', err=True)
        return out

    data = {'video_name': video_name, 'release_batch': release_batch}
    r = session.post('video/name', json=data)
    out = handle_request_error(r)

    if r.ok:
        v = out['response']
        click.echo(f'id={v["id"]} created for name={video_name}', err=True)
        click.echo(f'sending name={video_name} id={v["id"]} to S3', err=True)
        # TODO: upload to s3
        return out
    else:
        click.echo(f'name={video_name} creation failed', err=True)
        return out


@click.group(name='video', short_help='ingest and query video')
@click.pass_obj
def video(ctx):
    pass


@video.command(name='status', help='status of video annotation')
@click.argument('name', type=click.STRING)
@click.pass_obj
def status(ctx, name):
    r = ctx['session'].get('video', params={'name': name})
    if not r.ok and r.status_code != 404:
        exit_with(handle_request_error(r))
    elif r.status_code == 404 or not r.json():
        click.echo('Video not found, has it been ingested?', err=True)
        sys.exit(1)

    video_id = r.json()[0]['id']
    r = ctx['session'].get(f'video/{video_id}/status')
    headers = ['video', 'activity type', 'status']
    table = []

    for row in r.json()['vat_statuses']:
        table.append(
            [r.json()['video_name'], row['video_activity_type']['activity_type'], row['status'],]
        )
    # sort by activity type
    click.echo(
        tabulate(sorted(table, key=lambda row: row[1]), headers, tablefmt='github')  # type:ignore
    )


@video.command(name='search', help='search for video')
@click.option('--name', type=click.STRING)
@click.option('--gtag', type=click.STRING)
@click.option('--location', type=click.Choice([e.value for e in CameraLocation]))
@click.option('--release-batch', type=click.Choice([e.value for e in ReleaseBatches]))
@click.option('--scenario', type=click.Choice([e.value for e in Scenarios]))
@click.option('--page', type=click.INT)
@click.option('--size', type=click.INT)
@click.pass_obj
def search(ctx, **kwargs):
    data = {}
    for key, value in kwargs.items():
        if value is not None:
            data[key] = value
    r = ctx['session'].get('video', params=data)
    exit_with(handle_request_error(r))


@video.command(name='add', help='ingest video into stumpf from file')
@click.option(
    '--video-path', type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@click.option(
    '--release-batch', type=click.Choice([e.value for e in ReleaseBatches]), required=True
)
@click.pass_obj
def add(ctx, video_path, release_batch):
    exit_with(ingest_video(video_path, release_batch, ctx['session']))


cli.add_command(video)



from math import inf
import re
from typing import Callable, List

from boiler import BoilerError, BoilerWarning
from boiler.definitions import activity_spec, ActivityType, actor_codes
from boiler.models import Activity, ActivityList, Actor, Detection

# Ideally, we set this higher, but theoretically there could be a 3 frame track...
MAX_KEYFRAME_RATIO = 0.75


def validate_activity_actor_types(actor_string: str, activity_type: ActivityType) -> bool:
    pattern = activity_spec.get(activity_type)
    if pattern is not None:
        sorted_actor_string = ''.join(sorted(actor_string))
        pattern = f'^{pattern}$'
        result = re.match(pattern, sorted_actor_string)
        if result is None:
            raise BoilerWarning(
                (
                    f'activity spec validation failed for {activity_type.value}:'
                    f' {sorted_actor_string} found, {activity_spec[activity_type]} expected'
                )
            )
    return True


def validate_detection(detection: Detection):
    """
    * detections have 4 int corners
    * detection: top < bottom, left < right
    """
    box = detection.box
    if not (box.left < box.right):
        raise BoilerError('left ({box.left}) should be < right ({box.right})')
    if not (box.top < box.bottom):
        raise BoilerError('top ({box.top}) should be < bottom ({box.bottom})')


def validate_actor(actor: Actor, detection_validator: Callable = validate_detection):
    keyframe_count = 0
    last = None
    for detection in actor.detections:
        if last and detection.frame < last.frame:  # type: ignore
            raise BoilerError(
                f'detection_frame={detection.frame} was not greater than {last.frame}'
            )
        last = detection
        try:
            detection_validator(detection)
        except BoilerError as err:
            raise BoilerError(f'detection_frame={detection.frame}') from err
        if detection.keyframe:
            keyframe_count += 1
    if not (len(actor.detections) >= 2):
        raise BoilerError('actor must have at least a start and end frame')

    # TODO: add interpolated frame detection https://gitlab.com/diva-mturk/stumpf-diva/issues/30
    # failed_frame = actor.pruned()
    # assert failed_frame < 0, f'frame={failed_frame} interpolation detected on keyframe'

    if not actor.detections[0].frame <= actor.begin:
        raise BoilerError("actor's first detection must be <= actor begin frame")
    if not (actor.detections[0].keyframe):
        raise BoilerError('first detection must be keyframe')
    if not (actor.detections[-1].frame >= actor.end):
        raise BoilerError("actor's final detection must must be >= actor end frame")
    if not (actor.detections[-1].keyframe):
        raise BoilerError('last detection must be keyframe')

    total_frames = actor.detections[-1].frame - actor.detections[0].frame + 1
    if not (keyframe_count <= MAX_KEYFRAME_RATIO * total_frames):
        raise BoilerError(
            f'keyframe density {round(keyframe_count / total_frames * 100)}%'
            f' higher than {round(MAX_KEYFRAME_RATIO * 100)}%'
        )


def validate_activity(
    activity: Activity,
    actor_validator: Callable = validate_actor,
    detection_validator: Callable = validate_detection,
):
    """
    * actors involved are appropriate for activity type
    * actors framerange is within activity framerange
    """
    actor_string = ''
    activity_type = ActivityType(activity.activity_type)
    first_actor_frame = inf
    last_actor_frame = 0
    for actor in activity.actors:
        try:
            actor_validator(actor, detection_validator=detection_validator)
        except BoilerError as err:
            raise BoilerError(f'actor_id={actor.actor_id}') from err
        actor_type = actor.actor_type
        actor_string += actor_codes[actor_type]
        first_actor_frame = min(first_actor_frame, actor.begin)
        last_actor_frame = max(last_actor_frame, actor.end)
        if actor.begin > activity.end:
            raise BoilerError('actor track must begin before activity ends')
        if actor.end < activity.begin:
            raise BoilerError('actor track must end after activity begins')
    if first_actor_frame > activity.begin:
        raise BoilerError('there must be an actor at the first frame of an activity')
    if last_actor_frame < activity.end:
        raise BoilerError('there must be an actor at the last frame of an activity')
    validate_activity_actor_types(actor_string, activity_type)


def validate_activities(
    activity_list: ActivityList,
    activity_validator: Callable = validate_activity,
    actor_validator: Callable = validate_actor,
    detection_validator: Callable = validate_detection,
):
    warnings: List[Exception] = []
    fatals: List[Exception] = []
    for activity in activity_list:
        try:
            activity_validator(
                activity, actor_validator=actor_validator, detection_validator=detection_validator,
            )
        except BoilerWarning as err:
            new_err = BoilerError(f'activity_id={activity.activity_id}')
            new_err.__cause__ = err
            warnings.append(new_err)
        except Exception as err:
            new_err = BoilerError(f'activity_id={activity.activity_id}')
            new_err.__cause__ = err
            fatals.append(new_err)
    return warnings, fatals



import datetime
import os
import platform
import sys
import traceback
from typing import Optional


import click
from packaging.version import parse as parse_version
from pkg_resources import DistributionNotFound, get_distribution
import requests
from requests.exceptions import RequestException
from requests_toolbelt.sessions import BaseUrlSession
from xdg import BaseDirectory

__version__ = None
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass

BOILER_CONFIG_PATH = __name__
BOILER_CREDENTIAL_FILE = 'credentials'


class BoilerException(Exception):
    pass


class BoilerWarning(BoilerException):
    pass


class BoilerError(BoilerException):
    pass


class BoilerSession(BaseUrlSession):

    page_size = 50

    def __init__(self, base_url, token):
        base_url = f'{base_url.rstrip("/")}/'  # tolerate input with or without trailing slash
        super(BoilerSession, self).__init__(base_url=base_url)
        self.token = token
        self.headers.update(
            {
                'User-agent': f'boiler/{__version__}',
                'Accept': 'application/json',
                'X-Stumpf-Token': self.token,
            }
        )

    def request(self, *args, **kwargs):
        response = super().request(*args, **kwargs)

        if response.status_code in [401, 403]:
            click.echo(
                click.style(
                    "You are attempting to perform an authorized operation but you aren't logged in.\n"  # noqa
                    'Run the following command: boiler login',
                    fg='yellow',
                ),
                err=True,
            )
            sys.exit(1)

        return response


def newer_version_available():
    if __version__ is None:
        return False

    this_version = parse_version(__version__)
    r = requests.get('https://pypi.org/pypi/diva-boiler/json', timeout=(5, 5))
    r.raise_for_status()
    releases = [parse_version(v) for v in r.json()['releases'].keys()]
    for release in releases:
        if not (release.is_prerelease or release.is_devrelease) and release > this_version:
            return True
    return False


def main():
    try:
        cli()
    except Exception:
        click.echo(
            click.style(
                'The following unexpected error occurred while attempting your operation:\n',
                fg='red',
            ),
            err=True,
        )

        click.echo(traceback.format_exc(), err=True)

        click.echo(f'boiler:  v{__version__}', err=True)
        click.echo(f'python:  v{platform.python_version()}', err=True)
        click.echo(f'time:    {datetime.datetime.now(datetime.timezone.utc).isoformat()}', err=True)
        click.echo(f'os:      {platform.platform()}', err=True)
        click.echo(f'command: {" ".join(sys.argv[1:])}\n', err=True)

        click.echo(
            click.style(
                'This is a bug in boiler and should be reported here with the above output:',
                fg='yellow',
            ),
            err=True,
        )
        click.echo(
            'https://gitlab.com/diva-mturk/stumpf-diva/issues/new', err=True,
        )


def _get_boiler_token() -> Optional[str]:
    config_dir = BaseDirectory.load_first_config(BOILER_CONFIG_PATH)

    if config_dir:
        credential_file = os.path.join(config_dir, BOILER_CREDENTIAL_FILE)
        if os.path.exists(credential_file):
            return open(credential_file, 'r').read()
    return None


@click.group()
@click.option(
    '--api-url', default='https://stumpf.avidannotations.com/api/diva/', envvar='STUMPF_API_URL',
)
@click.option('--x-stumpf-token', envvar='X_STUMPF_TOKEN', default=_get_boiler_token)
@click.option('--offline', is_flag=True)
@click.version_option()
@click.pass_context
def cli(ctx, api_url, x_stumpf_token, offline):
    if not offline:
        try:
            if newer_version_available():
                click.echo(
                    click.style(
                        """There is a newer version of boiler available.
You must upgrade to the latest version before continuing.
If you are using pipx, then you can upgrade by running the following command:
""",
                        fg='yellow',
                    ),
                    err=True,
                )
                click.echo(click.style('pipx upgrade diva-boiler', fg='green'), err=True)
                sys.exit(1)
        except RequestException:
            click.echo(
                click.style('Failed to check for newer version of boiler:', fg='red'), err=True
            )
            raise

    # make boiler configuration directory
    BaseDirectory.save_config_path(BOILER_CONFIG_PATH)

    session = BoilerSession(api_url, x_stumpf_token)
    ctx.obj = {'session': session, 'stumpf_url': api_url.replace('/api/diva', '').rstrip('/')}


# TODO: re-enable kpf once deserialization is fixed
from boiler.commands import (  # noqa: F401 E402
    activity,
    export,
    gunrunner,
    kw18,
    login,
    vendor,
    video,
)



from typing import Any, Dict, IO, List

import click
from requests import Session

from boiler import cli
from boiler.commands.utils import (
    exit_with,
    handle_request_error,
    summarize_activities,
)
from boiler.serialization import (
    kpf as kpf_serialization,
    kw18 as kw18_serialization,
    web as web_serialization,
)
from boiler.validate import validate_activities

serializers = {'kw18': kw18_serialization}


class KPFIngestError(Exception):
    pass


def _upload_kpf_component(session: Session, file: IO[bytes]) -> Dict[str, Any]:
    file.seek(0)
    resp = session.post('upload', files={'file': file})
    file.seek(0)
    if resp.status_code == 409:
        # for these purposes, we don't care if it was a conflict or not
        resp.status_code = 201

    return handle_request_error(resp)


def _ingest_single_video(
    session: Session,
    types: IO[bytes],
    geom: IO[bytes],
    activities: IO[bytes],
    video_name: str,
    vendor_name: str,
    activity_type_list: List[str],
):
    types_id = _upload_kpf_component(session, types)
    geom_id = _upload_kpf_component(session, geom)
    activities_id = _upload_kpf_component(session, activities)

    data = {
        'activity_types': activity_type_list,
        'kpf_activities_id': activities_id,
        'kpf_geom_id': geom_id,
        'kpf_types_id': types_id,
        'validate_existing_files': True,  # set to False for faster no-op processing
        'vendor_name': vendor_name,
        'video_name': video_name,
    }
    resp = session.post('video-pipeline/process-annotation', json=data)
    return handle_request_error(resp)


@click.group(name='kpf', short_help='Ingest and validate kpf')
@click.pass_obj
def kpf(ctx):
    pass


@kpf.command(name='load', help='interact locally with the KPF format')
@click.option('--types', type=click.File(mode='rb'), help='path to types.yml')
@click.option('--geom', type=click.File(mode='rb'), help='path to geom.yml')
@click.option('--activities', type=click.File(mode='rb'), help='path to activities.yml')
@click.option('--validate', is_flag=True, help='run static integrity checks')
@click.option('--prune', is_flag=True, help='prune interpolated detections before validation')
@click.option(
    '--convert',
    nargs=2,
    type=click.Tuple(
        [
            click.Choice(serializers.keys()),
            click.Path(
                exists=False, file_okay=True, dir_okay=False, writable=True, resolve_path=True
            ),
        ]
    ),
    default=(None, None),
    help='convert to another serialization',
)
def load(types, geom, activities, validate, prune, convert):
    # TODO: fix
    raise NotImplementedError('This command is currently broken, sorry')

    actor_map = {}  # type: ignore
    activity_map = {}
    if types:
        kpf_serialization.deserialize_types(types, actor_map)
    if activities:
        kpf_serialization.deserialize_activities(activities, activity_map, actor_map)
    if geom:
        kpf_serialization.deserialize_geom(geom, actor_map)
    if prune:
        [a.prune() for a in activity_map.values()]

    output = {'summary': summarize_activities(list(activity_map.values()))}

    if validate:
        warnings, errors = validate_activities(activity_map.values())
        output['error'] = web_serialization.serialize_validation_errors(errors)
        output['warning'] = web_serialization.serialize_validation_errors(warnings)

    serialization_type = convert[0]
    base_path = convert[1]
    if serialization_type and base_path:
        serializer = serializers[serialization_type]
        serializer.serialize_to_files(
            base_path, activity_map.values(), actor_map.values(), keyframes_only=prune
        )

    exit_with(output)


cli.add_command(kpf)



from typing import List, TextIO

import click

from boiler import cli
from boiler.commands.kw18 import _find_kw18_files, _video_name_regexp
from boiler.commands.utils import activity_types_from_file
from boiler.serialization.kw18 import KW18FileSet


@click.group(name='gunrunner', short_help='Manage transitions to gunrunner')
@click.pass_obj
def gunrunner(ctx):
    pass


@gunrunner.command(name='dispatch', short_help='Dispatch a set of activities to gunrunner')
@click.argument('path', type=click.Path(exists=True, file_okay=False, resolve_path=True), nargs=-1)
@click.option('--activity-type-list', type=click.File(mode='r'), required=True)
@click.option('--recursive', '-r', is_flag=True)
@click.pass_context
def dispatch(ctx, path: List[str], recursive: bool, activity_type_list: TextIO):
    """Dispatch one or more videos to gunrunner.

    This command is intended to be run from the audit repository.  Because the
    list of activities from annotation -> audit is not necessarily the same,
    the user must provide the correct activity type list.  (Get the current
    list of activities with `boiler activity list-types`.)

    \b
    Example:
    boiler gunrunner dispatch -r --activity-type-list 47-activities.txt \\
        m2-annotations-audit/2018-10-15/08/
    """
    session = ctx.obj['session']
    has_error = False
    activity_types = activity_types_from_file(activity_type_list)

    for base_path in _find_kw18_files(path, recursive):
        video_name_match = _video_name_regexp.search(str(base_path))
        if video_name_match is None:
            has_error = True
            click.echo(f'Could not determine video name from "{base_path}"')
            continue

        video_name = video_name_match.group()

        click.echo(f'Dispatching {video_name}...')
        kw18_file_upload = KW18FileSet.create_from_path(base_path).upload(session)

        data = {
            'video_name': video_name,
            'activity_types': activity_types,
            'kw18_file_upload_id': kw18_file_upload['id'],
        }
        resp = session.post('video-pipeline/gunrunner/dispatch', json=data)
        try:
            response_data = resp.json()
        except Exception:
            response_data = {}

        if not resp.ok:
            message = response_data.get('error', 'Unknown error')
            click.secho(f'Failed to dispatch {video_name}', fg='red')
            click.echo(f'=> {message}')
            continue

        click.secho(f'Successfully dispatched {video_name}', fg='green')
    ctx.exit(0 if not has_error else 1)


cli.add_command(gunrunner)



from typing import Dict

import pytest

from boiler import BoilerError
from boiler.definitions import ActorType
from boiler.models import Activity, Actor, Box
from boiler.serialization import kpf


def test_deserialize_types_ex1():
    actor_map_ex1: Dict[int, Actor] = {}
    input_ex1 = """
    - { meta: 'foo' }
    - { meta: { team: 'Kitware' } }
    """
    kpf.deserialize_types(input_ex1, actor_map_ex1)
    assert len(actor_map_ex1.values()) == 0


@pytest.mark.skip('Fix immmutable models')
def test_deserialize_types_ex2():
    actor_map_ex2: Dict[int, Actor] = {}
    input_ex2 = """
    - { types: { id1: 1, cset3: { Person: 1.0 } } }
    - { types: { id1: 2, cset3: { Vehicle: 1.0 } } }
    - { types: { id1: 2, cset3: { Horse: 1.0 } } }
    """
    kpf.deserialize_types(input_ex2, actor_map_ex2)
    assert len(actor_map_ex2.values()) == 2
    assert 1 in actor_map_ex2 and 2 in actor_map_ex2
    assert actor_map_ex2[1].actor_type == ActorType.PERSON.value  # type: ignore


def test_deserialize_types_ex3():
    actor_map_ex3: Dict[int, Actor] = {}
    input_ex3 = """
    - { types: { id1: 18, cset3: { Bag: 0.5, Person: 0.5 } } }
    """
    with pytest.raises(BoilerError) as err:
        kpf.deserialize_types(input_ex3, actor_map_ex3)
    assert 'cset3' in str(err.value)


@pytest.mark.skip('Fix immmutable models')
def test_deserialize_geom():
    actor_map_ex1: Dict[int, Actor] = {}
    input_ex1 = """
    - { meta: '2018-05-18.13-05-03.13-10-02.school.G336' }
    - { meta: { team: 'Kitware' } }
    - { geom: {id1: 1, ts0: 2, g0: 615 575 943 703, keyframe: true } }
    - { geom: {id1: 1, ts0: 1, g0: 20 897 934 510 } }
    - { foo: 'bar' }
    - { geom: {id1: 2, ts0: 2, g0: 70 865 926 530 } }
    - { geom: {id1: 2, ts0: 3, g0: 119 832 917 550 } }
    - { }
    - { geom: {id1: 2, ts0: 1, g0: 515 574 850 710, keyframe: true } }
    """
    kpf.deserialize_geom(input_ex1, actor_map_ex1)
    assert len(actor_map_ex1.values()) == 2
    assert 1 in actor_map_ex1 and 2 in actor_map_ex1
    assert len(actor_map_ex1[1].detections) == 2
    assert len(actor_map_ex1[2].detections) == 3
    assert actor_map_ex1[1].detections[0].frame == 1
    assert actor_map_ex1[2].detections[0].keyframe is True
    assert actor_map_ex1[1].detections[0].box == Box(left=20, top=897, right=934, bottom=510)


@pytest.mark.skip('Fix immmutable models')
def test_deserialize_activities():
    actor_map_ex1: Dict[int, Actor] = {}
    activity_map_ex1: Dict[int, Activity] = {}
    input_ex1 = (
        '- {act: {act2: {Vehicle_Moving: 1.0}, id2: 1, timespan: [{tsr0: [1, 100]}], actors: [{id1: 1, timespan: [{tsr0: [2508, 2751]}]}] } }\n'  # noqa: E501
        '- {meta: "Vehicle_PicksUp_Person 1 instances"}\n'
        '- {act: {act2: {Vehicle_Moving: 1.0}, id2: 2, timespan: [{tsr0: [200, 300]}], src_status: audited, actors: [{id1: 2, timespan: [{tsr0: [2869, 3015]}]}]}}'  # noqa: E501
    )
    kpf.deserialize_activities(input_ex1, activity_map_ex1, actor_map_ex1)
    assert len(activity_map_ex1.values()) == 2
    assert len(activity_map_ex1[1].actors) == 1
    assert activity_map_ex1[1].begin == 1
    assert activity_map_ex1[2].end == 300
